@@ -1,11 +1,13 @@
 #! perl
 # 	$rcs = ' $Id: testout.t,v 1.2 1997/09/22 10:13:37 ilya Exp ilya $ ' ;
 
+use strict;
 use Math::Pari qw(:DEFAULT pari_print :all);
 use vars qw($x $y $z $k $t $q $a $u $j $l $name $other $n);
 die "Need a path to a testout file" unless @ARGV;
 
-$file = CORE::shift;
+my $file = CORE::shift;
+my(@tests, %seen, $current_num, $userfun, $installed, $skip_gnuplot, %not_yet_defined, $printout, $intests, $rest, $popped);
 {
   open TO, "< $file" or die "open `$file': $!";
   local $/ = "\n? ";
@@ -13,29 +15,52 @@ $file = CORE::shift;
   close TO or die "close: $!";
 }
 
-$mess = CORE::shift @tests unless @tests and $tests[0] =~ /^\?/; # Messages
-pop @tests;			# \q
-if ($tests = @tests) {
+my $mess = '';
+$mess = CORE::shift @tests if @tests and $tests[0] !~ /^\?/; # Messages
+
+(my $ifile = $file) =~ s[(?<=/test/)(32|64)/][in/];
+{
+  open FROM, '<', $ifile or die "open `$file': $!";
+  local $/;
+  $intests = <FROM>;
+  close FROM or die "close: $!";
+}
+(my $pref = $intests) =~ s/^(\\e|default\(echo,[^\n]*\n)(.*)//ms and $rest = $2;
+
+$popped++, pop @tests if ($tests[-1] || 0) =~ /^\\q/;
+
+if ($pref =~ /^\s*\\\\\s*package:(.*)/m) {
+  print "1..0 # skipped: package dependency: $1\n";
+  @tests = ();
+} elsif (my $tests = @tests) {
   print "1..$tests\n";
-} else {
-  print "1..0 # skipped: no tests found in `$file'\n";
+} elsif ($rest) {
+  print "1..1\nnot ok 1 # no tests found in `$file', but input has echo enabled\n";
+  warn "unexpected format of GP/PARI test output file: echo enabled, but no echoed lines";
+} else {	# no echoed lines expected
+  print "1..0 # skipped: command-line echoing is not enabled in the test file `$ifile'\n";
 }
 
-prec($3 || $1, 1) if ($mess || '') =~ /realprecision = (\d+) significant digits( \((\d+) digits displayed\))?/;
+# prec($3 || $1, 1) if $mess =~ /.*realprecision = (\d+) significant digits( \((\d+) digits displayed\))?/; # take the latest one
+parse_as_gp($pref) if $rest;
 
 #Math::Pari::dumpStack,allocatemem(8e6),Math::Pari::dumpStack if $file =~ /intnum/; # Due to our allocation policy we need more?
 
 
-$| = 1;
-@seen = qw(Euler Pi I getrand a x xx y z k t q u j l n v p e
-	   name other mhbi a2 a1 a0 b0 b1
-	   acurve bcurve ccurve cmcurve tcurve mcurve ma mpoints);
+$| = 1;		# Need to list loop variables used by the code:
+my @seen = qw(Euler Pi I getrand a x xx y z k t q u j l n v p e s
+	      name other mhbi a2 a1 a0 b0 b1
+	      acurve bcurve ccurve cmcurve tcurve mcurve ma mpoints);
+my @VARS = map "\$$_", @seen;
+eval 'use vars @VARS; 1' or die "use vars: $@";
 @seen{@seen}  = (' ', ' ', ' ', ' ', ('$') x 100);
+$seen{oo} = ' ' if Math::Pari::pari_version_exp >= 2009000;
 for (@seen) {
+  no strict 'refs';
   $$_ = PARI($_) unless $seen{$_} eq ' ';
 }
 $seen{'random'} = ' ';
-$DEFAULT = undef;
+my $DEFAULT = undef;
 
 # Some of these are repeated below (look for XXXX), since they cause
 # an early interpretation of unquoted args
@@ -63,16 +88,16 @@ if ($file =~ /plot|graph|all/) {
   }
 }
 
-$started = 0;
+my $started = 0;
 
 main_loop:
 for (@tests) {
   #print "Doing `$_'";
   1 while s/^\\\\.*\n//;	# Comment
-  $bad = /^\\/;			# \precision =
-  $wasbadprint = /\b(plot)\b/;
-  $wasprint = /\b((|p|tex)print(tex)?|plot)\b/;
-  s/\s*\n\?\s*\Z// or die "Not terminated: `$_'\n";
+  my $bad = /^\\/;			# \precision =
+  my $wasbadprint = /\b(plot)\b/;
+  my $wasprint = /\b((|p|tex)print(tex)?|plot)\b/;
+  s/\s*\n\?\s*\Z// or not $popped or die "Not terminated: `$_'\n";
   s/\A(\s*\?)+\s*//;
 #  s/[^\S\n]+$//gm;
   s/\A(.*)\s*; $ \s*(\w+)\s*\(/$1;$2(/mx; # Continuation lines of questions
@@ -80,7 +105,7 @@ for (@tests) {
   s/\A(p2=.*\d{10})\n(7\n)/$1$2/;
   s/\n(?=\/\w)//;
   s/\A(.*)\s*$//m or die "No question: `$_'\n";
-  $in = $1;
+  my $in = $1;
   1 while s/^\n//;		# Before warnings
   1 while s/^\s*\*+\s*warning.*\n?//i; # skip warnings
   if (s/^\s*\*+\s*(.*)//) {		# error
@@ -97,7 +122,7 @@ for (@tests) {
 #      chomp;
 #      push @ans, $_;
 #    }
-    @ans = split "\n";
+    my @ans = split "\n";
     if ($wasbadprint) {
       process_print($in, @ans);
     } elsif ($wasprint) {
@@ -171,7 +196,7 @@ sub mformat_transp {
 
 sub massage_floats {
   my $in = CORE::shift;
-  $pre = CORE::shift || "16g";
+  my $pre = CORE::shift || "16g";
   $in =~ s/(.\d*)\s+e/$1E/gi;	# 1.74 E-78
   $in =~ s/\b(\d+\.\d*(e[-+]?\d+)?|\d{10,})\b/sprintf "%.${pre}", $1/gei;
   $in;
@@ -196,8 +221,31 @@ sub process_cond {
   }
 }
 
+sub nok_print {
+  my ($n, $in) = (CORE::shift, CORE::shift);
+  print(@_), return unless $ENV{AUTOMATED_TESTING};
+  warn("# in = `$in'\n", @_);
+  print("not ok $n\n");
+}
+
+sub update_seen ($) {
+  my $seen_now = CORE::shift;
+  @seen{keys %$seen_now} = values %$seen_now;
+#  my @VARS = map "\$$_", keys %$seen_now;
+#  eval 'use vars @VARS; 1' or die "use vars: $@";
+}
+
+sub pre_update_seen ($) {
+  my $sym = CORE::shift;
+#  @seen{keys %$seen_now} = values %$seen_now;
+#  my @VARS = map "\$$_", keys %$seen_now;
+  eval "use vars '\$$sym'; 1" or die "use vars: $@";
+}
+
+my $prev;
 sub process_test {
   my ($in, $noans, $out) = @_;
+  my($res, $rres, $rout);
   my $ini_time = time;
   my $doprint;
   $doprint = 1 if $noans eq 'print';
@@ -372,6 +420,7 @@ sub process_test {
     }
     if ($in =~ /^\s*alias\s*\(\s*(\w+)\s*,\s*(\w+)\s*\)$/) {
       print "# Aliasing `$1' ==> `$2'\nok $current_num\n";
+      no strict 'refs';
       *$1 = \&{$2};
       return;
     }
@@ -397,7 +446,7 @@ sub process_test {
       # Substitute division
 	$in =~ s/(^|[\-\(,\[])(\d+)\s*\/\s*(\d+)(?=$|[\),\]])/$1 PARI($2)\/PARI($3) /g;
     }
-    %seen_now = ();
+    my %seen_now;
     # Substitute i= in loop commands
     if ($in !~ /\b(hermite|mathnf|until)\s*\(/) { # Special case, not loop-with-=
       $in =~ s/([\(,])(\w+)=(?!=)/ $seen_now{$2} = '$'; "$1$2," /ge;
@@ -407,12 +456,12 @@ sub process_test {
     $in =~ s/\b(|p|tex)print1\(/ 'my_' . $1 . 'print(0,'/ge;
     $in =~ s/\b(eval|shift|sort)\(/&$1\(/g; # eval($y)
     # Special case -oo (with $oo=[PARI(1)] done earlier;
-    # Having $oo defined without external PARI tests conversions; but it'sn't overloaded
-    $in =~ s/-oo\b/- PARI(\$oo)/ if $seen_now{oo} or 1;
+    # Having $oo defined without external PARI tests conversions; but it'sn't overloaded in older PARI
+    $in =~ s/-oo\b/- PARI(\$oo)/ if $seen_now{oo} or Math::Pari::pari_version_exp < 2009000;
     # Recognize variables and prepend $ in assignments
     # s/\b(direuler\s*\(\s*\w+\s*),/$1=/;	# direuler
     $in =~ s/\bX\b/PARIvar("X")/g if $in =~ /\bdireuler\b/;
-    $in =~ s/(^\s*|[;(]\s*)(?=(\w+)\s*=\s*)/$seen_now{$2} = '$'; $1 . '$'/ge; # Assignment
+    $in =~ s/(^\s*|[;(]\s*)(?=(\w+)\s*=\s*)/$seen_now{$2} = '$'; pre_update_seen($2); $1 . '$'/ge; # Assignment
     # Prepend $ to variables (not before '^' - inside of 'o(x^17)'):
     $in =~ s/(^|[^\$])\b([a-zA-Z]\w*)\b(?!\s*[\(^])/
       		($1 || '') . ($seen{$2} || $seen_now{$2} || '') . $2
@@ -701,7 +750,7 @@ sub process_test {
 	  and $not_yet_defined{$1}) {
 	print "# in='$in'\nok $current_num # Skipped: `$1' is known to be undefined\n";
       } elsif ($@ =~ /high resolution graphics disabled/
-	       and not Math::Pari::have_graphics()) {
+	       and 0 >= Math::Pari::have_graphics()) {
 	print "# in='$in'\nok $current_num # Skipped: graphic is disabled in this build\n";
       } elsif ($@ =~ /gnuplot-like plotting environment not loaded yet/
 	       and $skip_gnuplot) {
@@ -716,7 +765,7 @@ sub process_test {
       $cmp = eval { $rres =~ /^$re_out$/ };
       if ($@ and $@ =~ /regexp too big/) {
 	print "ok $current_num # Skipped: $@\n";
-	@seen{keys %seen_now} = values %seen_now;
+	update_seen \%seen_now;
 	$prev = $res;
 	return;
       }
@@ -729,7 +778,7 @@ sub process_test {
       "'\n# pari==='", join("\t", @$out), "'\n# re_out='$re_out'\n";
     } elsif (not $noans and defined $re_out) {
       print "ok $current_num #  run_time=$run_time, post_time=$post_time, pre_time=$pre_time\n";
-      @seen{keys %seen_now} = values %seen_now;
+      update_seen \%seen_now;
       $prev = $res;
     } elsif (not $noans and (not defined $rres or $rres ne $rout)) {
       nok_print $current_num, $in, "not ok $current_num # in='$in'\n#    out='", $rres, "', type='", ref $res,
@@ -739,19 +788,12 @@ sub process_test {
       "'\n#   expect='$rout', type='", ref $res,"'\n";
     } else {
       print "ok $current_num #  run_time=$run_time, post_time=$post_time, pre_time=$pre_time\n";
-      @seen{keys %seen_now} = values %seen_now;
+      update_seen \%seen_now;
       $prev = $res;
     }
   }
 }
 
-sub nok_print {
-  my ($n, $in) = (CORE::shift, CORE::shift);
-  print(@_), return unless $ENV{AUTOMATED_TESTING};
-  warn("# in = `$in'\n", @_);
-  print("not ok $n\n");
-}
-
 sub process_error {
   my ($in, $out, $error) = @_;
   $current_num++;
@@ -764,15 +806,15 @@ sub process_definition {
   eval "PARI('$def');  import Math::Pari $name;";
   if ($@) {
     chomp $@;
-    print("not ok $current_num # definition: `$in' error `$@'\n");
+    print("not ok $current_num # definition: `$def' error `$@'\n");
   } else {
-    print("# definition $current_num: `$in'\nok $current_num\n");
+    print("# definition $current_num: `$def'\nok $current_num\n");
   }
 }
 
 sub process_set {
   my ($in, $out) = @_;
-  return process_test("setprecision($1)", 'noans', '') if $in =~ /^\\p\s*(\d+)$/;
+  return process_test("setprecision($1)", 'noans', []) if $in =~ /^\\p\s*(\d+)$/;
   $current_num++;
   print("# `$in'\nok $current_num # Skipping setting test\n");
 }
